@@ -20,6 +20,7 @@
 #include <sys/queue.h>
 #include <sys/wait.h>
 #include <sys/mman.h>
+#include <sys/time.h>
 
 #include <openssl/pem.h>
 #include <openssl/x509v3.h>
@@ -40,11 +41,12 @@
 
 #include "kore.h"
 
-#if defined(OpenBSD) || defined(__FreeBSD_version)
+#if defined(OpenBSD) || defined(__FreeBSD_version) || \
+    defined(NetBSD) || defined(__DragonFly_version)
 #define PRI_TIME_T		"d"
 #endif
 
-#if defined(linux)
+#if defined(__linux__)
 #if defined(__x86_64__)
 #define PRI_TIME_T		PRIu64
 #else
@@ -56,7 +58,26 @@
 #define PRI_TIME_T		"ld"
 #endif
 
-#define LD_FLAGS_MAX		10
+#define LD_FLAGS_MAX		30
+#define CFLAGS_MAX		30
+#define CXXFLAGS_MAX		CFLAGS_MAX
+
+#define BUILD_NOBUILD		0
+#define BUILD_C			1
+#define BUILD_CXX		2
+
+struct buildopt {
+	char			*name;
+	char			*kore_source;
+	char			*kore_flavor;
+	int			single_binary;
+	struct kore_buf		*cflags;
+	struct kore_buf		*cxxflags;
+	struct kore_buf		*ldflags;
+	TAILQ_ENTRY(buildopt)	list;
+};
+
+TAILQ_HEAD(buildopt_list, buildopt);
 
 struct cmd {
 	const char		*name;
@@ -71,7 +92,6 @@ struct filegen {
 struct cfile {
 	struct stat		st;
 	int			build;
-	int			cpp;
 	char			*name;
 	char			*fpath;
 	char			*opath;
@@ -81,16 +101,19 @@ struct cfile {
 TAILQ_HEAD(cfile_list, cfile);
 
 static void		cli_fatal(const char *, ...) __attribute__((noreturn));
-
 static void		cli_file_close(int);
-static void		cli_run_kore(void *);
+static void		cli_run_kore(void);
 static void		cli_generate_certs(void);
 static void		cli_link_library(void *);
-static void		cli_compile_cfile(void *);
+static void		cli_compile_kore(void *);
+static void		cli_compile_source_file(void *);
 static void		cli_mkdir(const char *, int);
 static int		cli_dir_exists(const char *);
 static int		cli_file_exists(const char *);
 static void		cli_cleanup_files(const char *);
+static void		cli_build_cflags(struct buildopt *);
+static void		cli_build_cxxflags(struct buildopt *);
+static void		cli_build_ldflags(struct buildopt *);
 static void		cli_file_writef(int, const char *, ...);
 static void		cli_file_open(const char *, int, int *);
 static void		cli_file_remove(char *, struct dirent *);
@@ -99,19 +122,39 @@ static void		cli_file_write(int, const void *, size_t);
 static int		cli_vasprintf(char **, const char *, ...);
 static void		cli_spawn_proc(void (*cb)(void *), void *);
 static void		cli_write_asset(const char *, const char *);
-static void		cli_register_cfile(char *, struct dirent *);
+static void		cli_register_kore_file(char *, struct dirent *);
+static void		cli_register_source_file(char *, struct dirent *);
 static void		cli_file_create(const char *, const char *, size_t);
 static int		cli_file_requires_build(struct stat *, const char *);
 static void		cli_find_files(const char *,
 			    void (*cb)(char *, struct dirent *));
-static void		cli_add_cfile(char *, char *, char *,
-			    struct stat *, int, int);
+static void		cli_add_source_file(char *, char *, char *,
+			    struct stat *, int);
+
+static struct buildopt	*cli_buildopt_default(void);
+static struct buildopt	*cli_buildopt_new(const char *);
+static struct buildopt	*cli_buildopt_find(const char *);
+static void		cli_buildopt_cleanup(void);
+static void		cli_buildopt_parse(const char *);
+static void		cli_buildopt_cflags(struct buildopt *, const char *);
+static void		cli_buildopt_cxxflags(struct buildopt *, const char *);
+static void		cli_buildopt_ldflags(struct buildopt *, const char *);
+static void		cli_buildopt_single_binary(struct buildopt *,
+			    const char *);
+static void		cli_buildopt_kore_source(struct buildopt *,
+			    const char *);
+static void		cli_buildopt_kore_flavor(struct buildopt *,
+			    const char *);
+
+static void		cli_flavor_load(void);
+static void		cli_flavor_change(const char *);
 
 static void		cli_run(int, char **);
 static void		cli_help(int, char **);
 static void		cli_build(int, char **);
 static void		cli_clean(int, char **);
 static void		cli_create(int, char **);
+static void		cli_flavor(int, char **);
 
 static void		file_create_src(void);
 static void		file_create_config(void);
@@ -123,6 +166,7 @@ static struct cmd cmds[] = {
 	{ "build",	"build an application",			cli_build },
 	{ "clean",	"cleanup the build files",		cli_clean },
 	{ "create",	"create a new application skeleton",	cli_create },
+	{ "flavor",	"switch build flavor",			cli_flavor },
 	{ NULL,		NULL,					NULL }
 };
 
@@ -135,7 +179,7 @@ static struct filegen gen_files[] = {
 
 static const char *gen_dirs[] = {
 	"src",
-#if !defined(KORE_BENCHMARK)
+#if !defined(KORE_NO_TLS)
 	"cert",
 #endif
 	"conf",
@@ -157,23 +201,55 @@ static const char *src_data =
 	"}\n";
 
 static const char *config_data =
-	"# Placeholder configuration\n"
+	"# %s configuration\n"
 	"\n"
 	"bind\t\t127.0.0.1 8888\n"
 	"load\t\t./%s.so\n"
-#if !defined(KORE_BENCHMARK)
+#if !defined(KORE_NO_TLS)
 	"tls_dhparam\tdh2048.pem\n"
 #endif
 	"\n"
-	"domain 127.0.0.1 {\n"
-#if !defined(KORE_BENCHMARK)
+	"domain * {\n"
+#if !defined(KORE_NO_TLS)
 	"\tcertfile\tcert/server.crt\n"
 	"\tcertkey\t\tcert/server.key\n"
 #endif
 	"\tstatic\t/\tpage\n"
 	"}\n";
 
-#if !defined(KORE_BENCHMARK)
+static const char *build_data =
+	"# %s build config\n"
+	"# You can switch flavors using: kore flavor [newflavor]\n"
+	"\n"
+	"# Set to yes if you wish to produce a single binary instead\n"
+	"# of a dynamic library. If you set this to yes you must also\n"
+	"# set kore_source together with kore_flavor and update ldflags\n"
+	"# to include the appropriate libraries you will be linking with.\n"
+	"#single_binary=no\n"
+	"#kore_source=/home/joris/src/kore\n"
+	"#kore_flavor=\n"
+	"\n"
+	"# The flags below are shared between flavors\n"
+	"cflags=-Wall -Wmissing-declarations -Wshadow\n"
+	"cflags=-Wstrict-prototypes -Wmissing-prototypes\n"
+	"cflags=-Wpointer-arith -Wcast-qual -Wsign-compare\n"
+	"\n"
+	"cxxflags=-Wall -Wmissing-declarations -Wshadow\n"
+	"cxxflags=-Wpointer-arith -Wcast-qual -Wsign-compare\n"
+	"\n"
+	"dev {\n"
+	"	# These flags are added to the shared ones when\n"
+	"	# you build the \"dev\" flavor.\n"
+	"	cflags=-g\n"
+	"	cxxflags=-g\n"
+	"}\n"
+	"\n"
+	"#prod {\n"
+	"#	You can specify additional flags here which are only\n"
+	"#	included if you build with the \"prod\" flavor.\n"
+	"#}\n";
+
+#if !defined(KORE_NO_TLS)
 static const char *dh2048_data =
 	"-----BEGIN DH PARAMETERS-----\n"
 	"MIIBCAKCAQEAn4f4Qn5SudFjEYPWTbUaOTLUH85YWmmPFW1+b5bRa9ygr+1wfamv\n"
@@ -185,15 +261,27 @@ static const char *dh2048_data =
 	"-----END DH PARAMETERS-----";
 #endif
 
-static const char *gitignore_data = "*.o\n.objs\n%s.so\nassets.h\ncert\n";
+static const char *gitignore = "*.o\n.flavor\n.objs\n%s.so\nassets.h\ncert\n";
 
 static int			s_fd = -1;
 static char			*appl = NULL;
+static int			run_after = 0;
 static char			*rootdir = NULL;
-static char			*compiler = "gcc";
+static char			*compiler_c = "gcc";
+static char			*compiler_cpp = "g++";
+static char			*compiler_ld = "gcc";
 static struct cfile_list	source_files;
-static int			cfiles_count;
+static struct buildopt_list	build_options;
+static int			source_files_count;
+static int			cxx_files_count;
 static struct cmd		*command = NULL;
+static int			cflags_count = 0;
+static int			cxxflags_count = 0;
+static int			ldflags_count = 0;
+static char			*flavor = NULL;
+static char			*cflags[CFLAGS_MAX];
+static char			*cxxflags[CXXFLAGS_MAX];
+static char			*ldflags[LD_FLAGS_MAX];
 
 void
 kore_cli_usage(int local)
@@ -227,6 +315,9 @@ kore_cli_main(int argc, char **argv)
 
 	(void)umask(S_IWGRP|S_IWOTH);
 
+	if ((flavor = strchr(argv[0], ':')) != NULL)
+		*(flavor)++ = '\0';
+
 	for (i = 0; cmds[i].name != NULL; i++) {
 		if (!strcmp(argv[0], cmds[i].name)) {
 			argc--;
@@ -275,81 +366,162 @@ cli_create(int argc, char **argv)
 
 	cli_generate_certs();
 
-	printf("%s created succesfully!\n", appl);
+	printf("%s created successfully!\n", appl);
 
-#if !defined(KORE_BENCHMARK)
+#if !defined(KORE_NO_TLS)
 	printf("note: do NOT use the created DH parameters/certificates in production\n");
 #endif
 }
 
 static void
-cli_build(int argc, char **argv)
+cli_flavor(int argc, char **argv)
 {
-	struct cfile	*cf;
-	struct timeval	times[2];
-	int		requires_relink;
-	char		pwd[PATH_MAX], *src_path, *assets_header;
-	char		*assets_path, *p, *obj_path, *cpath, *config;
+	struct buildopt		*bopt;
+	char			pwd[MAXPATHLEN], *conf;
+
+	if (getcwd(pwd, sizeof(pwd)) == NULL)
+		cli_fatal("could not get cwd: %s", errno_s);
+
+	appl = basename(pwd);
+	(void)cli_vasprintf(&conf, "conf/%s.conf", appl);
+	if (!cli_dir_exists("conf") || !cli_file_exists(conf))
+		cli_fatal("%s doesn't appear to be a kore app", appl);
+	free(conf);
+
+	TAILQ_INIT(&build_options);
+	(void)cli_buildopt_new("_default");
+	cli_buildopt_parse("conf/build.conf");
 
 	if (argc == 0) {
-		if (getcwd(pwd, sizeof(pwd)) == NULL)
-			cli_fatal("could not get cwd: %s", errno_s);
-
-		rootdir = ".";
-		appl = basename(pwd);
+		cli_flavor_load();
+		TAILQ_FOREACH(bopt, &build_options, list) {
+			if (!strcmp(bopt->name, "_default"))
+				continue;
+			if (!strcmp(bopt->name, flavor)) {
+				printf("* %s\n", bopt->name);
+			} else {
+				printf("  %s\n", bopt->name);
+			}
+		}
 	} else {
-		appl = argv[0];
-		rootdir = appl;
+		cli_flavor_change(argv[0]);
+		printf("changed build flavor to: %s\n", argv[0]);
 	}
 
-	if ((p = getenv("CC")) != NULL)
-		compiler = p;
+	cli_buildopt_cleanup();
+}
 
-	cfiles_count = 0;
+static void
+cli_build(int argc, char **argv)
+{
+	struct dirent		dp;
+	struct cfile		*cf;
+	struct buildopt		*bopt;
+	struct timeval		times[2];
+	char			*build_path;
+	int			requires_relink, l;
+	char			*sofile, *config, *data;
+	char			*assets_path, *p, *obj_path, *cpath;
+	char			pwd[PATH_MAX], *src_path, *assets_header;
+
+	if (getcwd(pwd, sizeof(pwd)) == NULL)
+		cli_fatal("could not get cwd: %s", errno_s);
+
+	rootdir = ".";
+	appl = basename(pwd);
+
+	if ((p = getenv("CC")) != NULL)
+		compiler_c = p;
+
+	if ((p = getenv("CXX")) != NULL)
+		compiler_cpp = p;
+
+	source_files_count = 0;
+	cxx_files_count = 0;
 	TAILQ_INIT(&source_files);
+	TAILQ_INIT(&build_options);
 
 	(void)cli_vasprintf(&src_path, "%s/src", rootdir);
 	(void)cli_vasprintf(&assets_path, "%s/assets", rootdir);
 	(void)cli_vasprintf(&config, "%s/conf/%s.conf", rootdir, appl);
 	(void)cli_vasprintf(&assets_header, "%s/src/assets.h", rootdir);
+	(void)cli_vasprintf(&build_path, "%s/conf/build.conf", rootdir);
+
 	if (!cli_dir_exists(src_path) || !cli_file_exists(config))
 		cli_fatal("%s doesn't appear to be a kore app", appl);
 
-	free(config);
+	cli_flavor_load();
+	bopt = cli_buildopt_new("_default");
+	if (!cli_file_exists(build_path)) {
+		l = cli_vasprintf(&data, build_data, appl);
+		cli_file_create("conf/build.conf", data, l);
+		free(data);
+	}
+
+	cli_find_files(src_path, cli_register_source_file);
+	free(src_path);
+
+	cli_buildopt_parse(build_path);
+	free(build_path);
 
 	(void)cli_vasprintf(&obj_path, "%s/.objs", rootdir);
 	if (!cli_dir_exists(obj_path))
 		cli_mkdir(obj_path, 0755);
 	free(obj_path);
 
+	if (bopt->single_binary) {
+		if (bopt->kore_source == NULL)
+			cli_fatal("single_binary set but not kore_source");
+
+		printf("building kore (%s)\n", bopt->kore_source);
+		cli_spawn_proc(cli_compile_kore, bopt);
+
+		(void)cli_vasprintf(&src_path, "%s/src", bopt->kore_source);
+		cli_find_files(src_path, cli_register_kore_file);
+		free(src_path);
+	}
+
+	printf("building %s (%s)\n", appl, flavor);
+
+	cli_build_cflags(bopt);
+	cli_build_cxxflags(bopt);
+	cli_build_ldflags(bopt);
+
 	(void)unlink(assets_header);
 
 	/* Generate the assets. */
-	if (cli_dir_exists(assets_path)) {
-		cli_file_open(assets_header,
-		    O_CREAT | O_TRUNC | O_WRONLY, &s_fd);
+	cli_file_open(assets_header, O_CREAT | O_TRUNC | O_WRONLY, &s_fd);
+	cli_file_writef(s_fd, "#ifndef __H_KORE_ASSETS_H\n");
+	cli_file_writef(s_fd, "#define __H_KORE_ASSETS_H\n");
 
-		cli_file_writef(s_fd, "#ifndef __H_KORE_ASSETS_H\n");
-		cli_file_writef(s_fd, "#define __H_KORE_ASSETS_H\n");
+	if (cli_dir_exists(assets_path))
 		cli_find_files(assets_path, cli_build_asset);
-		cli_file_writef(s_fd, "\n#endif\n");
-		cli_file_close(s_fd);
+
+	if (bopt->single_binary) {
+		memset(&dp, 0, sizeof(dp));
+		dp.d_type = DT_REG;
+		printf("adding config %s\n", config);
+		(void)snprintf(dp.d_name,
+		    sizeof(dp.d_name), "builtin_kore.conf");
+		cli_build_asset(config, &dp);
 	}
 
+	cli_file_writef(s_fd, "\n#endif\n");
+	cli_file_close(s_fd);
+
 	free(assets_path);
+	free(config);
+
+	if (cxx_files_count > 0)
+		compiler_ld = compiler_cpp;
 
-	/* Build all source files. */
-	cli_find_files(src_path, cli_register_cfile);
-	
-	free(src_path);
 	requires_relink = 0;
-
 	TAILQ_FOREACH(cf, &source_files, list) {
-		if (cf->build == 0)
+		if (cf->build == BUILD_NOBUILD)
 			continue;
 
 		printf("compiling %s\n", cf->name);
-		cli_spawn_proc(cli_compile_cfile, cf);
+		cli_spawn_proc(cli_compile_source_file, cf);
 
 		times[0].tv_usec = 0;
 		times[0].tv_sec = cf->st.st_mtime;
@@ -371,12 +543,26 @@ cli_build(int argc, char **argv)
 	}
 	free(cpath);
 
-	if (requires_relink) {
-		cli_spawn_proc(cli_link_library, NULL);
-		printf("%s built succesfully!\n", appl);
+	if (bopt->single_binary) {
+		requires_relink++;
+		(void)cli_vasprintf(&sofile, "%s", appl);
 	} else {
-		printf("nothing to be done\n");
+		(void)cli_vasprintf(&sofile, "%s.so", appl);
 	}
+
+	if (!cli_file_exists(sofile))
+		requires_relink++;
+	free(sofile);
+
+	if (requires_relink) {
+		cli_spawn_proc(cli_link_library, bopt);
+		printf("%s built successfully!\n", appl);
+	} else {
+		printf("nothing to be done!\n");
+	}
+
+	if (run_after == 0)
+		cli_buildopt_cleanup();
 }
 
 static void
@@ -401,6 +587,7 @@ cli_clean(int argc, char **argv)
 static void
 cli_run(int argc, char **argv)
 {
+	run_after = 1;
 	cli_build(argc, argv);
 
 	if (chdir(rootdir) == -1)
@@ -410,7 +597,7 @@ cli_run(int argc, char **argv)
 	 * We are exec()'ing kore again, while we could technically set
 	 * the right cli options manually and just continue running.
 	 */
-	cli_run_kore(NULL);
+	cli_run_kore();
 }
 
 static void
@@ -430,11 +617,14 @@ file_create_config(void)
 	char		*name, *data;
 
 	(void)cli_vasprintf(&name, "conf/%s.conf", appl);
-	l = cli_vasprintf(&data, config_data, appl);
+	l = cli_vasprintf(&data, config_data, appl, appl);
 	cli_file_create(name, data, l);
-
 	free(name);
 	free(data);
+
+	l = cli_vasprintf(&data, build_data, appl);
+	cli_file_create("conf/build.conf", data, l);
+	free(data);
 }
 
 static void
@@ -443,7 +633,7 @@ file_create_gitignore(void)
 	int		l;
 	char		*data;
 
-	l = cli_vasprintf(&data, gitignore_data, appl);
+	l = cli_vasprintf(&data, gitignore, appl);
 	cli_file_create(".gitignore", data, l);
 	free(data);
 }
@@ -600,6 +790,12 @@ cli_build_asset(char *fpath, struct dirent *dp)
 	if (stat(fpath, &st) == -1)
 		cli_fatal("stat: %s %s", fpath, errno_s);
 
+	/* If this file was empty, skip it. */
+	if (st.st_size == 0) {
+		printf("skipping empty asset %s\n", name);
+		return;
+	}
+
 	(void)cli_vasprintf(&opath, "%s/.objs/%s.o", rootdir, name);
 	(void)cli_vasprintf(&cpath, "%s/.objs/%s.c", rootdir, name);
 
@@ -608,9 +804,9 @@ cli_build_asset(char *fpath, struct dirent *dp)
 		*(ext)++ = '\0';
 		cli_write_asset(name, ext);
 		*ext = '_';
-		
-		cli_add_cfile(name, cpath, opath, &st, 0, 0);
-		kore_mem_free(name);
+
+		cli_add_source_file(name, cpath, opath, &st, BUILD_NOBUILD);
+		kore_free(name);
 		return;
 	}
 
@@ -631,7 +827,7 @@ cli_build_asset(char *fpath, struct dirent *dp)
 
 	/* Start generating the file. */
 	cli_file_writef(out, "/* Auto generated */\n");
-	cli_file_writef(out, "#include <sys/param.h>\n\n");
+	cli_file_writef(out, "#include <sys/types.h>\n\n");
 
 	/* Write the file data as a byte array. */
 	cli_file_writef(out, "u_int8_t asset_%s_%s[] = {\n", name, ext);
@@ -668,22 +864,21 @@ cli_build_asset(char *fpath, struct dirent *dp)
 	*--ext = '.';
 
 	/* Register the .c file now (cpath is free'd later). */
-	cli_add_cfile(name, cpath, opath, &st, 1, 0);
-	kore_mem_free(name);
+	cli_add_source_file(name, cpath, opath, &st, BUILD_C);
+	kore_free(name);
 }
 
 static void
-cli_add_cfile(char *name, char *fpath, char *opath, struct stat *st,
-    int build, int cpp)
+cli_add_source_file(char *name, char *fpath, char *opath, struct stat *st,
+    int build)
 {
 	struct cfile		*cf;
 
-	cfiles_count++;
+	source_files_count++;
 	cf = kore_malloc(sizeof(*cf));
 
 	cf->st = *st;
 	cf->build = build;
-	cf->cpp = cpp;
 	cf->fpath = fpath;
 	cf->opath = opath;
 	cf->name = kore_strdup(name);
@@ -692,31 +887,59 @@ cli_add_cfile(char *name, char *fpath, char *opath, struct stat *st,
 }
 
 static void
-cli_register_cfile(char *fpath, struct dirent *dp)
+cli_register_source_file(char *fpath, struct dirent *dp)
 {
 	struct stat		st;
 	char			*ext, *opath;
-	int			cpp;
+	int			build;
 
 	if ((ext = strrchr(fpath, '.')) == NULL ||
 	    (strcmp(ext, ".c") && strcmp(ext, ".cpp")))
 		return;
 
+	if (stat(fpath, &st) == -1)
+		cli_fatal("stat(%s): %s", fpath, errno_s);
+
 	if (!strcmp(ext, ".cpp"))
-		cpp = 1;
-	else
-		cpp = 0;
+		cxx_files_count++;
+
+	(void)cli_vasprintf(&opath, "%s/.objs/%s.o", rootdir, dp->d_name);
+	if (!cli_file_requires_build(&st, opath)) {
+		build = BUILD_NOBUILD;
+	} else if (!strcmp(ext, ".cpp")) {
+		build = BUILD_CXX;
+	} else {
+		build = BUILD_C;
+	}
+
+	cli_add_source_file(dp->d_name, fpath, opath, &st, build);
+}
+
+static void
+cli_register_kore_file(char *fpath, struct dirent *dp)
+{
+	struct stat		st, ost;
+	char			*opath, *ext, *fname;
+
+	if ((ext = strrchr(fpath, '.')) == NULL || strcmp(ext, ".c"))
+		return;
 
 	if (stat(fpath, &st) == -1)
 		cli_fatal("stat(%s): %s", fpath, errno_s);
 
-	(void)cli_vasprintf(&opath, "%s/.objs/%s.o", rootdir, dp->d_name);
-	if (!cli_file_requires_build(&st, opath)) {
-		cli_add_cfile(dp->d_name, fpath, opath, &st, 0, cpp);
+	*ext = '\0';
+	if ((fname = basename(fpath)) == NULL)
+		cli_fatal("basename failed");
+
+	(void)cli_vasprintf(&opath, "%s/.objs/%s.o", rootdir, fname);
+
+	/* Silently ignore non existing object files for kore source files. */
+	if (stat(opath, &ost) == -1) {
+		free(opath);
 		return;
 	}
 
-	cli_add_cfile(dp->d_name, fpath, opath, &st, 1, cpp);
+	cli_add_source_file(dp->d_name, fpath, opath, &st, BUILD_NOBUILD);
 }
 
 static void
@@ -766,7 +989,7 @@ cli_find_files(const char *path, void (*cb)(char *, struct dirent *))
 static void
 cli_generate_certs(void)
 {
-#if !defined(KORE_BENCHMARK)
+#if !defined(KORE_NO_TLS)
 	BIGNUM			*e;
 	FILE			*fp;
 	time_t			now;
@@ -866,59 +1089,37 @@ cli_generate_certs(void)
 }
 
 static void
-cli_compile_cfile(void *arg)
+cli_compile_source_file(void *arg)
 {
-	int		idx;
+	int		idx, i;
 	struct cfile	*cf = arg;
-	char		*args[24], *ipath[2], *p, *cppstandard;
-#if defined(KORE_USE_PGSQL)
-	char		*ppath;
-#endif
+	char		*args[32 + CFLAGS_MAX];
+	char		*compiler;
+	char		**flags;
+	int		flags_count;
 
-	(void)cli_vasprintf(&ipath[0], "-I%s/src", rootdir);
-	(void)cli_vasprintf(&ipath[1], "-I%s/src/includes", rootdir);
+	switch (cf->build) {
+	case BUILD_C:
+		compiler = compiler_c;
+		flags = cflags;
+		flags_count = cflags_count;
+		break;
+	case BUILD_CXX:
+		compiler = compiler_cpp;
+		flags = cxxflags;
+		flags_count = cxxflags_count;
+		break;
+	default:
+		cli_fatal("cli_compile_file: unexpected file type: %d",
+		    cf->build);
+		break;
+	}
 
-	/*
-	 * These compiler options should be settable
-	 * somehow by the user if they so choose.
-	 */
 	idx = 0;
 	args[idx++] = compiler;
-	args[idx++] = ipath[0];
-	args[idx++] = ipath[1];
-#if defined(PREFIX)
-	(void)cli_vasprintf(&args[idx++], "-I%s/include", PREFIX);
-#else
-	args[idx++] = "-I/usr/local/include";
-#endif
 
-#if defined(KORE_USE_PGSQL)
-	(void)cli_vasprintf(&ppath, "-I%s", PGSQL_INCLUDE_PATH);
-	args[idx++] = ppath;
-#endif
-
-	args[idx++] = "-Wall";
-	args[idx++] = "-Wmissing-declarations";
-	args[idx++] = "-Wshadow";
-	args[idx++] = "-Wpointer-arith";
-	args[idx++] = "-Wcast-qual";
-	args[idx++] = "-Wsign-compare";
-	args[idx++] = "-fPIC";
-	args[idx++] = "-g";
-
-	if (cf->cpp) {
-		args[idx++] = "-Woverloaded-virtual";
-		args[idx++] = "-Wold-style-cast";
-		args[idx++] = "-Wnon-virtual-dtor";
-
-		if ((p = getenv("CXXSTD")) != NULL) {
-			(void)cli_vasprintf(&cppstandard, "-std=%s", p);
-			args[idx++] = cppstandard;
-		}
-	} else {
-		args[idx++] = "-Wstrict-prototypes";
-		args[idx++] = "-Wmissing-prototypes";
-	}
+	for (i = 0; i < flags_count; i++)
+		args[idx++] = flags[i];
 
 	args[idx++] = "-c";
 	args[idx++] = cf->fpath;
@@ -927,74 +1128,496 @@ cli_compile_cfile(void *arg)
 	args[idx] = NULL;
 
 	execvp(compiler, args);
+	cli_fatal("failed to start '%s': %s", compiler, errno_s);
 }
 
 static void
 cli_link_library(void *arg)
 {
 	struct cfile		*cf;
-	int			idx, f, i, has_cpp;
-	char			*args[cfiles_count + 11 + LD_FLAGS_MAX];
-	char			*p, *libname, *flags[LD_FLAGS_MAX], *cpplib;
+	struct buildopt		*bopt;
+	int			idx, i;
+	char			*output;
+	char			*args[source_files_count + 11 + LD_FLAGS_MAX];
 
-	if ((p = getenv("LDFLAGS")) != NULL)
-		f = kore_split_string(p, " ", flags, LD_FLAGS_MAX);
+	bopt = arg;
+
+	if (bopt->single_binary)
+		(void)cli_vasprintf(&output, "%s/%s", rootdir, appl);
 	else
-		f = 0;
-
-	(void)cli_vasprintf(&libname, "%s/%s.so", rootdir, appl);
+		(void)cli_vasprintf(&output, "%s/%s.so", rootdir, appl);
 
 	idx = 0;
-	args[idx++] = compiler;
+	args[idx++] = compiler_ld;
 
-#if defined(__MACH__)
-	args[idx++] = "-dynamiclib";
-	args[idx++] = "-undefined";
-	args[idx++] = "suppress";
-	args[idx++] = "-flat_namespace";
-#else
-	args[idx++] = "-shared";
-#endif
-
-	has_cpp = 0;
-	TAILQ_FOREACH(cf, &source_files, list) {
-		if (cf->cpp)
-			has_cpp = 1;
+	TAILQ_FOREACH(cf, &source_files, list)
 		args[idx++] = cf->opath;
-	}
 
-	if (has_cpp) {
-		if ((p = getenv("CXXLIB")) != NULL) {
-			(void)cli_vasprintf(&cpplib, "-l%s", p);
-			args[idx++] = cpplib;
-		} else {
-			args[idx++] = "-lstdc++";
-		}
-	}
+	for (i = 0; i < ldflags_count; i++)
+		args[idx++] = ldflags[i];
 
-	for (i = 0; i < f; i++)
-		args[idx++] = flags[i];
+	if (bopt->single_binary) {
+		args[idx++] = "-rdynamic";
+#if defined(__linux__)
+		args[idx++] = "-ldl";
+#endif
+	}
 
 	args[idx++] = "-o";
-	args[idx++] = libname;
+	args[idx++] = output;
 	args[idx] = NULL;
 
-	execvp(compiler, args);
+	execvp(compiler_ld, args);
+	cli_fatal("failed to start '%s': %s", compiler_ld, errno_s);
 }
 
 static void
-cli_run_kore(void *arg)
+cli_compile_kore(void *arg)
 {
-	char		*args[4], *cpath;
+	struct buildopt		*bopt = arg;
+	int			idx, i, fcnt;
+	char			*obj, *args[16], pwd[MAXPATHLEN], *flavors[7];
 
-	(void)cli_vasprintf(&cpath, "conf/%s.conf", appl);
+	if (getcwd(pwd, sizeof(pwd)) == NULL)
+		cli_fatal("could not get cwd: %s", errno_s);
 
-	args[0] = "kore";
-	args[1] = "-fnrc";
-	args[2] = cpath;
-	args[3] = NULL;
+	(void)cli_vasprintf(&obj, "OBJDIR=%s/.objs", pwd);
 
-	execvp("kore", args);
+	if (putenv(obj) != 0)
+		cli_fatal("cannot set OBJDIR for building kore");
+
+	if (putenv("CFLAGS=-DKORE_SINGLE_BINARY") != 0)
+		cli_fatal("cannot set CFLAGS for building kore");
+
+	fcnt = kore_split_string(bopt->kore_flavor, " ", flavors, 7);
+
+#if defined(OpenBSD) || defined(__FreeBSD_version) || \
+    defined(NetBSD) || defined(__DragonFly_version)
+	args[0] = "gmake";
+#else
+	args[0] = "make";
+#endif
+
+	args[1] = "-s";
+	args[2] = "-C";
+	args[3] = bopt->kore_source;
+	args[4] = "objects";
+
+	idx = 5;
+	for (i = 0; i < fcnt; i++) {
+		printf("using flavor %s\n", flavors[i]);
+		args[idx++] = flavors[i];
+	}
+
+	args[idx] = NULL;
+
+	execvp(args[0], args);
+	cli_fatal("failed to start '%s': %s", args[0], errno_s);
+}
+
+static void
+cli_run_kore(void)
+{
+	struct buildopt		*bopt;
+	char			*args[4], *cpath, *cmd, *flags;
+
+	bopt = cli_buildopt_default();
+
+	if (bopt->single_binary) {
+		cpath = NULL;
+		flags = "-fnr";
+		(void)cli_vasprintf(&cmd, "./%s", appl);
+	} else {
+		cmd = "kore";
+		flags = "-fnrc";
+		(void)cli_vasprintf(&cpath, "conf/%s.conf", appl);
+	}
+
+	args[0] = cmd;
+	args[1] = flags;
+
+	if (cpath != NULL) {
+		args[2] = cpath;
+		args[3] = NULL;
+	} else {
+		args[2] = NULL;
+	}
+
+	execvp(args[0], args);
+	cli_fatal("failed to start '%s': %s", args[0], errno_s);
+}
+
+static void
+cli_buildopt_parse(const char *path)
+{
+	FILE			*fp;
+	struct buildopt		*bopt;
+	char			buf[BUFSIZ], *p, *t;
+
+	if ((fp = fopen(path, "r")) == NULL)
+		cli_fatal("cli_buildopt_parse: fopen(%s): %s", path, errno_s);
+
+	bopt = NULL;
+
+	while ((p = kore_read_line(fp, buf, sizeof(buf))) != NULL) {
+		if (strlen(p) == 0)
+			continue;
+
+		if (bopt != NULL && !strcmp(p, "}")) {
+			bopt = NULL;
+			continue;
+		}
+
+		if (bopt == NULL) {
+			if ((t = strchr(p, '=')) != NULL)
+				goto parse_option;
+			if ((t = strchr(p, ' ')) == NULL)
+				cli_fatal("unexpected '%s'", p);
+			*(t)++ = '\0';
+			if (strcmp(t, "{"))
+				cli_fatal("expected '{', got '%s'", t);
+			bopt = cli_buildopt_new(p);
+			continue;
+		}
+
+		if ((t = strchr(p, '=')) == NULL) {
+			printf("bad buildopt line: '%s'\n", p);
+			continue;
+		}
+
+parse_option:
+		*(t)++ = '\0';
+
+		p = kore_text_trim(p, strlen(p));
+		t = kore_text_trim(t, strlen(t));
+
+		if (!strcasecmp(p, "cflags")) {
+			cli_buildopt_cflags(bopt, t);
+		} else if (!strcasecmp(p, "cxxflags")) {
+			cli_buildopt_cxxflags(bopt, t);
+		} else if (!strcasecmp(p, "ldflags")) {
+			cli_buildopt_ldflags(bopt, t);
+		} else if (!strcasecmp(p, "single_binary")) {
+			cli_buildopt_single_binary(bopt, t);
+		} else if (!strcasecmp(p, "kore_source")) {
+			cli_buildopt_kore_source(bopt, t);
+		} else if (!strcasecmp(p, "kore_flavor")) {
+			cli_buildopt_kore_flavor(bopt, t);
+		} else {
+			printf("ignoring unknown option '%s'\n", p);
+		}
+	}
+}
+
+static struct buildopt *
+cli_buildopt_new(const char *name)
+{
+	struct buildopt		*bopt;
+
+	bopt = kore_malloc(sizeof(*bopt));
+	bopt->cflags = NULL;
+	bopt->cxxflags = NULL;
+	bopt->ldflags = NULL;
+	bopt->single_binary = 0;
+	bopt->kore_source = NULL;
+	bopt->kore_flavor = NULL;
+	bopt->name = kore_strdup(name);
+
+	TAILQ_INSERT_TAIL(&build_options, bopt, list);
+	return (bopt);
+}
+
+static struct buildopt *
+cli_buildopt_find(const char *name)
+{
+	struct buildopt		*bopt;
+
+	TAILQ_FOREACH(bopt, &build_options, list) {
+		if (!strcmp(bopt->name, name))
+			return (bopt);
+	}
+
+	return (NULL);
+}
+
+static struct buildopt *
+cli_buildopt_default(void)
+{
+	struct buildopt		*bopt;
+
+	if ((bopt = cli_buildopt_find("_default")) == NULL)
+		fatal("no _default buildopt options");
+
+	return (bopt);
+}
+
+static void
+cli_buildopt_cleanup(void)
+{
+	struct buildopt		*bopt, *next;
+
+	for (bopt = TAILQ_FIRST(&build_options); bopt != NULL; bopt = next) {
+		next = TAILQ_NEXT(bopt, list);
+		TAILQ_REMOVE(&build_options, bopt, list);
+
+		if (bopt->cflags != NULL)
+			kore_buf_free(bopt->cflags);
+		if (bopt->cxxflags != NULL)
+			kore_buf_free(bopt->cxxflags);
+		if (bopt->ldflags != NULL)
+			kore_buf_free(bopt->ldflags);
+		if (bopt->kore_source != NULL)
+			kore_free(bopt->kore_source);
+		if (bopt->kore_flavor != NULL)
+			kore_free(bopt->kore_flavor);
+		kore_free(bopt);
+	}
+}
+
+static void
+cli_buildopt_cflags(struct buildopt *bopt, const char *string)
+{
+	if (bopt == NULL)
+		bopt = cli_buildopt_default();
+
+	if (bopt->cflags == NULL)
+		bopt->cflags = kore_buf_alloc(128);
+
+	kore_buf_appendf(bopt->cflags, "%s ", string);
+}
+
+static void
+cli_buildopt_cxxflags(struct buildopt *bopt, const char *string)
+{
+	if (bopt == NULL)
+		bopt = cli_buildopt_default();
+
+	if (bopt->cxxflags == NULL)
+		bopt->cxxflags = kore_buf_alloc(128);
+
+	kore_buf_appendf(bopt->cxxflags, "%s ", string);
+}
+
+static void
+cli_buildopt_ldflags(struct buildopt *bopt, const char *string)
+{
+	if (bopt == NULL)
+		bopt = cli_buildopt_default();
+
+	if (bopt->ldflags == NULL)
+		bopt->ldflags = kore_buf_alloc(128);
+
+	kore_buf_appendf(bopt->ldflags, "%s ", string);
+}
+
+static void
+cli_buildopt_single_binary(struct buildopt *bopt, const char *string)
+{
+	if (bopt == NULL)
+		bopt = cli_buildopt_default();
+	else
+		cli_fatal("single_binary only supported in global context");
+
+	if (!strcmp(string, "yes"))
+		bopt->single_binary = 1;
+	else
+		bopt->single_binary = 0;
+}
+
+static void
+cli_buildopt_kore_source(struct buildopt *bopt, const char *string)
+{
+	if (bopt == NULL)
+		bopt = cli_buildopt_default();
+	else
+		cli_fatal("kore_source only supported in global context");
+
+	if (bopt->kore_source != NULL)
+		kore_free(bopt->kore_source);
+
+	bopt->kore_source = kore_strdup(string);
+}
+
+static void
+cli_buildopt_kore_flavor(struct buildopt *bopt, const char *string)
+{
+	if (bopt == NULL)
+		bopt = cli_buildopt_default();
+	else
+		cli_fatal("kore_flavor only supported in global context");
+
+	if (bopt->kore_flavor != NULL)
+		kore_free(bopt->kore_flavor);
+
+	bopt->kore_flavor = kore_strdup(string);
+}
+
+static void
+cli_build_flags_common(struct kore_buf* buf)
+{
+	kore_buf_appendf(buf,
+	    "-fPIC -I%s/src -I%s/src/includes ", rootdir, rootdir);
+#if defined(PREFIX)
+	kore_buf_appendf(buf, "-I%s/include ", PREFIX);
+#else
+	kore_buf_appendf(buf, "-I/usr/local/include ");
+#endif
+#if defined(__MACH__)
+	/* Add default openssl include path from homebrew / ports under OSX. */
+	kore_buf_appendf(buf, "-I/opt/local/include ");
+	kore_buf_appendf(buf, "-I/usr/local/opt/openssl/include ");
+#endif
+#if defined(KORE_USE_PGSQL)
+	kore_buf_appendf(buf, "-I%s ", PGSQL_INCLUDE_PATH);
+#endif
+#if defined(KORE_NO_HTTP)
+	kore_buf_appendf(buf, "-DKORE_NO_HTTP ");
+#endif
+#if defined(KORE_NO_TLS)
+	kore_buf_appendf(buf, "-DKORE_NO_TLS ");
+#endif
+}
+
+static void
+cli_build_cflags(struct buildopt *bopt)
+{
+	struct buildopt		*obopt;
+	char			*string;
+
+	if ((obopt = cli_buildopt_find(flavor)) == NULL)
+		cli_fatal("no such build flavor: %s", flavor);
+
+	if (bopt->cflags == NULL)
+		bopt->cflags = kore_buf_alloc(128);
+
+	cli_build_flags_common(bopt->cflags);
+
+	if (obopt != NULL && obopt->cflags != NULL) {
+		kore_buf_append(bopt->cflags, obopt->cflags->data,
+		    obopt->cflags->offset);
+	}
+
+	if (bopt->single_binary)
+		kore_buf_appendf(bopt->cflags, "-DKORE_SINGLE_BINARY");
+
+	string = kore_buf_stringify(bopt->cflags, NULL);
+	printf("CFLAGS=%s\n", string);
+	cflags_count = kore_split_string(string, " ", cflags, CFLAGS_MAX);
+}
+
+static void
+cli_build_cxxflags(struct buildopt *bopt)
+{
+	struct buildopt		*obopt;
+	char			*string;
+
+	if ((obopt = cli_buildopt_find(flavor)) == NULL)
+		cli_fatal("no such build flavor: %s", flavor);
+
+	if (bopt->cxxflags == NULL)
+		bopt->cxxflags = kore_buf_alloc(128);
+
+	cli_build_flags_common(bopt->cxxflags);
+
+	if (obopt != NULL && obopt->cxxflags != NULL) {
+		kore_buf_append(bopt->cxxflags, obopt->cxxflags->data,
+		    obopt->cxxflags->offset);
+	}
+
+	string = kore_buf_stringify(bopt->cxxflags, NULL);
+	if (cxx_files_count > 0)
+		printf("CXXFLAGS=%s\n", string);
+	cxxflags_count = kore_split_string(string, " ", cxxflags, CXXFLAGS_MAX);
+}
+
+static void
+cli_build_ldflags(struct buildopt *bopt)
+{
+	struct buildopt		*obopt;
+	char			*string;
+
+	if ((obopt = cli_buildopt_find(flavor)) == NULL)
+		cli_fatal("no such build flavor: %s", flavor);
+
+	if (bopt->ldflags == NULL)
+		bopt->ldflags = kore_buf_alloc(128);
+
+	if (bopt->single_binary == 0) {
+#if defined(__MACH__)
+		kore_buf_appendf(bopt->ldflags,
+		    "-dynamiclib -undefined suppress -flat_namespace ");
+#else
+		kore_buf_appendf(bopt->ldflags, "-shared ");
+#endif
+	}
+
+	if (obopt != NULL && obopt->ldflags != NULL) {
+		kore_buf_append(bopt->ldflags, obopt->ldflags->data,
+		    obopt->ldflags->offset);
+	}
+
+	string = kore_buf_stringify(bopt->ldflags, NULL);
+	printf("LDFLAGS=%s\n", string);
+	ldflags_count = kore_split_string(string, " ", ldflags, LD_FLAGS_MAX);
+}
+
+static void
+cli_flavor_load(void)
+{
+	FILE		*fp;
+	char		buf[BUFSIZ], pwd[MAXPATHLEN], *p, *conf;
+
+	if (getcwd(pwd, sizeof(pwd)) == NULL)
+		cli_fatal("could not get cwd: %s", errno_s);
+
+	appl = basename(pwd);
+	if (appl == NULL)
+		cli_fatal("basename: %s", errno_s);
+	appl = kore_strdup(appl);
+	(void)cli_vasprintf(&conf, "conf/%s.conf", appl);
+
+	if (!cli_dir_exists("conf") || !cli_file_exists(conf))
+		cli_fatal("%s doesn't appear to be a kore app", appl);
+	free(conf);
+
+	if ((fp = fopen(".flavor", "r")) == NULL) {
+		flavor = kore_strdup("dev");
+		return;
+	}
+
+	if (fgets(buf, sizeof(buf), fp) == NULL)
+		cli_fatal("failed to read flavor from file");
+
+	if ((p = strchr(buf, '\n')) != NULL)
+		*p = '\0';
+
+	flavor = kore_strdup(buf);
+	(void)fclose(fp);
+}
+
+static void
+cli_flavor_change(const char *name)
+{
+	FILE			*fp;
+	int			ret;
+	struct buildopt		*bopt;
+
+	if ((bopt = cli_buildopt_find(name)) == NULL)
+		cli_fatal("no such flavor: %s", name);
+
+	if ((fp = fopen(".flavor.tmp", "w")) == NULL)
+		cli_fatal("failed to open temporary file to save flavor");
+
+	ret = fprintf(fp, "%s\n", name);
+	if (ret == -1 || (size_t)ret != (strlen(name) + 1))
+		cli_fatal("failed to write new build flavor");
+
+	(void)fclose(fp);
+
+	if (rename(".flavor.tmp", ".flavor") == -1)
+		cli_fatal("failed to replace build flavor");
+
+	cli_clean(0, NULL);
 }
 
 static void
